# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
from collections.abc import Callable

import numpy as np
import torch
from einops import rearrange

from anemoi.training.data.grid_indices import BaseGridIndices
from anemoi.training.utils.seeding import get_base_seed

from .singledataset import NativeGridDataset

LOGGER = logging.getLogger(__name__)


class EnsNativeGridDataset(NativeGridDataset):
    """Iterable ensemble dataset for AnemoI data on the arbitrary grids."""

    def __init__(
        self,
        data_reader: Callable,
        grid_indices: type[BaseGridIndices],
        relative_date_indices: list,
        timestep: str = "6h",
        shuffle: bool = True,
        label: str = "generic",
        ens_members_per_device: int = 1,
        num_gpus_per_ens: int = 1,
        num_gpus_per_model: int = 1,
    ) -> None:
        """Initialize (part of) the dataset state.

        Parameters
        ----------
        data_reader : Callable
            user function that opens and returns the zarr array data
        relative_date_indices: list
            list of time indices to load from the data relative to the current sample i in __iter__
        ens_members_per_device: int, optional
            number of ensemble members input for each GPU device, by default 1
        timestep: str, optional
            timestep of the data, by default "6h"
        shuffle : bool, optional
            Shuffle batches, by default True
        ens_members_per_device : int, optional
            Number of ensemble members input for each GPU device, by default 1
        num_gpus_per_ens : int, optional
            Number of GPUs per ensemble, by default 1
        num_gpus_per_model : int, optional
            Number of GPUs per model, by default 1

        Raises
        ------
        RuntimeError
            Multistep value cannot be negative.
        """
        super().__init__(
            data_reader=data_reader,
            relative_date_indices=relative_date_indices,
            timestep=timestep,
            shuffle=shuffle,
            grid_indices=grid_indices,
            label=label,
        )

        # Lazy init
        self.ens_members_per_device = ens_members_per_device
        self.num_gpus_per_ens = num_gpus_per_ens
        self.num_gpus_per_model = num_gpus_per_model

        # lazy init model and reader group info, will be set by the DDPGroupStrategy:
        self.ens_comm_group_rank = 0
        self.ens_comm_num_groups = 1
        self.ens_comm_group_id = 0

    @property
    def num_eda_members(self) -> int:
        """Return number of EDA members."""
        return self.data.shape[2] - 1

    @property
    def eda_flag(self) -> bool:
        """Return whether EDA is enabled."""
        return self.data.shape[2] > 1

    def sample_eda_members(self, num_eda_members: int = 9) -> np.ndarray:
        """Subselect EDA ensemble members assigned to the current device."""
        tot_ens = self.ens_members_per_device * self.num_gpus_per_ens // self.num_gpus_per_model

        assert (
            tot_ens <= num_eda_members
        ), f"Can't generate an ensemble of size {tot_ens} from {num_eda_members} EDA perturbations"

        eda_member_gen_idx = self.rng_inicond_sampling.choice(range(num_eda_members), size=tot_ens, replace=False)
        offset = 1  # index=0 analysis, index=1 EDA recentred
        eda_member_gen_idx += offset

        effective_rank = self.ens_comm_group_rank // self.num_gpus_per_model
        eda_member_idx = np.sort(
            eda_member_gen_idx[
                effective_rank * self.ens_members_per_device : self.ens_members_per_device * (1 + effective_rank)
            ],
        )

        LOGGER.debug(
            "GPU with global rank %s, comm_group_id %s, comm_group_rank %s will receive EDA member(s) %s",
            self.global_rank,
            self.ens_comm_group_id,
            self.ens_comm_group_rank,
            eda_member_gen_idx,
        )

        return eda_member_gen_idx, eda_member_idx

    def set_comm_group_info(
        self,
        global_rank: int,
        model_comm_group_id: int,
        model_comm_group_rank: int,
        model_comm_num_groups: int,
        ens_comm_group_id: int,
        ens_comm_group_rank: int,
        ens_comm_num_groups: int,
        reader_group_rank: int,
        reader_group_size: int,
    ) -> None:
        """Set model and reader communication group information (called by DDPGroupStrategy).

        Parameters
        ----------
        global_rank : int
            Global rank
        model_comm_group_id : int
            Model communication group ID
        model_comm_group_rank : int
            Model communication group rank
        model_comm_num_groups : int
            Number of model communication groups
        ens_comm_group_id : int
            Ensemble communication group ID
        ens_comm_group_rank : int
            Ensemble communication group rank
        ens_comm_num_groups : int
            Number of ensemble communication groups
        reader_group_rank : int
            Reader group rank
        reader_group_size : int
            Reader group size
        """
        self.global_rank = global_rank
        self.model_comm_group_id = model_comm_group_id
        self.model_comm_group_rank = model_comm_group_rank
        self.model_comm_num_groups = model_comm_num_groups
        self.ens_comm_group_id = ens_comm_group_id
        self.ens_comm_group_rank = ens_comm_group_rank
        self.ens_comm_num_groups = ens_comm_num_groups
        self.reader_group_rank = reader_group_rank
        self.reader_group_size = reader_group_size

        self.sample_comm_group_id = ens_comm_group_id  # groups that work on the same sample / batch
        self.sample_comm_num_groups = ens_comm_num_groups

        assert self.reader_group_size >= 1, "reader_group_size must be positive"

        LOGGER.info(
            "NativeGridDataset.set_group_info(): global_rank %d, model_comm_group_id %d, "
            "model_comm_group_rank %d, model_comm_num_groups %d, reader_group_rank %d",
            global_rank,
            model_comm_group_id,
            model_comm_group_rank,
            model_comm_num_groups,
            reader_group_rank,
        )

        LOGGER.info(
            "NativeGridDataset.set_group_info(): global_rank %d, ens_comm_group_id %d, "
            "ens_comm_group_rank %d, ens_comm_num_groups %d, reader_group_rank %d",
            global_rank,
            ens_comm_group_id,
            ens_comm_group_rank,
            ens_comm_num_groups,
            reader_group_rank,
        )

    def per_worker_init(self, n_workers: int, worker_id: int) -> None:
        """Called by worker_init_func on each copy of dataset.

        This initialises after the worker process has been spawned.

        Parameters
        ----------
        n_workers : int
            Number of workers
        worker_id : int
            Worker ID

        """
        super().per_worker_init(n_workers, worker_id)

        base_seed = get_base_seed()
        seed = (
            base_seed * (self.sample_comm_group_id + 1) - worker_id
        )  # note that test, validation etc. datasets get same seed
        self.rng_inicond_sampling = np.random.default_rng(seed=seed)
        sanity_rnd = self.rng.random(1)
        sanity_rnd_ini = self.rng_inicond_sampling.random(1)

        LOGGER.info(
            (
                "Worker %d (%s, pid %d, glob. rank %d, model comm group %d, "
                "model comm group rank %d, ens comm group %d, ens comm group rank %d, "
                " seed group id %d, seed %d, sanity rnd %f, sanity rnd ini %f)"
            ),
            worker_id,
            self.label,
            os.getpid(),
            self.global_rank,
            self.model_comm_group_id,
            self.model_comm_group_rank,
            self.ens_comm_group_id,
            self.ens_comm_group_rank,
            self.sample_comm_group_id,
            seed,
            sanity_rnd,
            sanity_rnd_ini,
        )

    def __iter__(self):
        """Return an iterator over the dataset.

        The datasets are retrieved by Anemoi Datasets from zarr files. This iterator yields
        chunked batches for DDP and sharded training.
        """
        if self.shuffle:
            shuffled_chunk_indices = self.rng.choice(
                self.valid_date_indices,
                size=len(self.valid_date_indices),
                replace=False,
            )[self.chunk_index_range]
        else:
            shuffled_chunk_indices = self.valid_date_indices[self.chunk_index_range]

        LOGGER.debug(
            (
                "Worker pid %d, label %s, worker id %d, global_rank %d, "
                "model comm group %d, group_rank %d, seed comm group id %d, using indices[0:10]: %s"
            ),
            os.getpid(),
            self.label,
            self.worker_id,
            self.global_rank,
            self.model_comm_group_id,
            self.model_comm_group_rank,
            self.sample_comm_group_id,
            shuffled_chunk_indices[:10],
        )

        for i in shuffled_chunk_indices:
            # start and end time indices, for analysis and EDA
            start = i + self.relative_date_indices[0]
            end_an = i + self.relative_date_indices[-1] + 1
            timeincrement = self.relative_date_indices[1] - self.relative_date_indices[0]
            # NOTE: this is temporary until anemoi datasets allows indexing with arrays or lists
            # data[start...] will be replaced with data[self.relative_date_indices + i]
            end_eda = i + timeincrement

            if self.eda_flag:
                _, eda_member_idx = self.sample_eda_members(self.num_eda_members)
            else:
                eda_member_idx = None

            grid_shard_indices = self.grid_indices.get_shard_indices(self.reader_group_rank)
            if isinstance(grid_shard_indices, slice):
                # Load only shards into CPU memory
                x_an = self.data[start:end_an:timeincrement, :, 0:1, grid_shard_indices]
            else:
                # Load full grid in CPU memory, select grid_shard after
                # Note that anemoi-datasets currently doesn't support slicing + indexing
                # in the same operation.
                x_an = self.data[start:end_an:timeincrement, :, 0:1, ...]
                x_an = x_an[..., grid_shard_indices]  # select the grid shard
            x_an = rearrange(x_an, "dates variables ensemble gridpoints -> dates ensemble gridpoints variables")

            x_pert: torch.Tensor | None = None
            if self.eda_flag:
                if isinstance(grid_shard_indices, slice):
                    x_pert = self.data[start : end_eda : self.timeincrement, ..., grid_shard_indices]
                else:
                    x_pert = self.data[start : end_eda : self.timeincrement, ...]
                    x_pert = x_pert[..., grid_shard_indices]  # select the grid shard
                x_pert = rearrange(
                    x_pert[:, :, eda_member_idx, ...],
                    "dates variables ensemble gridpoints -> dates ensemble gridpoints variables",
                )

            if x_pert is not None:
                sample = (
                    torch.from_numpy(x_an),
                    torch.from_numpy(x_pert),
                )
            else:
                sample = (torch.from_numpy(x_an),)

            yield sample

    def __repr__(self) -> str:
        return (
            f"""
            {super().__repr__()}
            Dataset: {self.data}
            Relative dates: {self.relative_date_indices}
            EDA: {self.eda_flag}
            """
            f"""
            Number of EDA members:
            {self.num_eda_members}" if self.eda_flag else
            """
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# ruff: noqa: ANN001


import logging
from pathlib import Path
from typing import Any

import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.utilities.types import STEP_OUTPUT

LOGGER = logging.getLogger(__name__)


class MemorySnapshotRecorder(Callback):
    """Record memory snapshot using torch.cuda._record_memory_history()."""

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.dirpath = Path(self.config.hardware.paths.profiler)

        self.warmup = self.config.diagnostics.benchmark_profiler.snapshot.warmup
        if not self.warmup:
            self.warmup = 0
        self.num_steps = (
            self.config.diagnostics.benchmark_profiler.snapshot.steps + self.warmup
        )  # be consistent with profiler scheduler
        self.status = False

        assert (
            self.num_steps % self.config.dataloader.batch_size.training == 0
        ), "Snapshot steps is not a multiple of batch size"
        assert (
            self.warmup % self.config.dataloader.batch_size.training == 0
        ), "Snapshot Warmup steps is not a multiple of batch size"

    @rank_zero_only
    def _start_snapshot_recording(self) -> None:
        LOGGER.info("Starting snapshot record_memory_history")
        torch.cuda.memory._record_memory_history()
        self.status = True

    @rank_zero_only
    def _save_snapshot(self) -> None:
        self.memory_snapshot_fname = self.dirpath / "memory_snapshot.pickle"
        try:
            LOGGER.info("Saving memory snapshot to %s", self.memory_snapshot_fname)
            torch.cuda.memory._dump_snapshot(f"{self.memory_snapshot_fname}")
        except BaseException:
            LOGGER.exception("Failed to capture memory snapshot")

    @rank_zero_only
    def stop_record_memory_history(self) -> None:
        LOGGER.info("Stopping snapshot record_memory_history")
        torch.cuda.memory._record_memory_history(enabled=None)

    def on_train_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Any,
        batch_idx: int,
    ) -> None:
        del pl_module, batch, batch_idx
        if trainer.global_step == self.warmup:
            self._start_snapshot_recording()

    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ) -> None:
        del batch, batch_idx, pl_module, outputs
        if trainer.global_step == self.num_steps:
            if self.status is True:
                self._save_snapshot()
                self.stop_record_memory_history()
            else:
                LOGGER.info("Snapshot recording was not started so no snapshot was saved")



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import sys
from typing import Any

from omegaconf import DictConfig
from omegaconf import OmegaConf
from pydantic import BaseModel as PydanticBaseModel
from pydantic import model_validator
from pydantic._internal import _model_construction
from pydantic_core import PydanticCustomError
from pydantic_core import ValidationError
from typing_extensions import Self

from anemoi.graphs.schemas.base_graph import BaseGraphSchema
from anemoi.models.schemas.decoder import GraphTransformerDecoderSchema
from anemoi.models.schemas.models import ModelSchema
from anemoi.utils.schemas import BaseModel
from anemoi.utils.schemas.errors import CUSTOM_MESSAGES
from anemoi.utils.schemas.errors import convert_errors

# to make these available at runtime for pydantic, bug should be resolved in
# future versions (see https://github.com/astral-sh/ruff/issues/7866)
from .data import DataSchema
from .dataloader import DataLoaderSchema
from .datamodule import DataModuleSchema
from .diagnostics import DiagnosticsSchema
from .hardware import HardwareSchema
from .training import TrainingSchema

_object_setattr = _model_construction.object_setattr

LOGGER = logging.getLogger(__name__)


class BaseSchema(BaseModel):
    """Top-level schema for the training configuration."""

    data: DataSchema
    """Data configuration."""
    dataloader: DataLoaderSchema
    """Dataloader configuration."""
    datamodule: DataModuleSchema
    """Datamodule configuration."""
    diagnostics: DiagnosticsSchema
    """Diagnostics configuration such as logging, plots and metrics."""
    hardware: HardwareSchema
    """Hardware configuration."""
    graph: BaseGraphSchema
    """Graph configuration."""
    model: ModelSchema
    """Model configuration."""
    training: TrainingSchema
    """Training configuration."""
    config_validation: bool = True
    """Flag to disable validation of the configuration"""

    @model_validator(mode="after")
    def set_read_group_size_if_not_provided(self) -> Self:
        if not self.dataloader.read_group_size:
            self.dataloader.read_group_size = self.hardware.num_gpus_per_model
        return self

    @model_validator(mode="after")
    def check_log_paths_available_for_loggers(self) -> Self:
        logger = []
        if self.diagnostics.log.wandb.enabled and (not self.hardware.paths.logs or not self.hardware.paths.logs.wandb):
            logger.append("wandb")
        if self.diagnostics.log.mlflow.enabled and (
            not self.hardware.paths.logs or not self.hardware.paths.logs.mlflow
        ):
            logger.append("mlflow")
        if self.diagnostics.log.tensorboard.enabled and (
            not self.hardware.paths.logs or not self.hardware.paths.logs.tensorboard
        ):
            logger.append("tensorboard")

        if logger:
            msg = ", ".join(logger) + " logging path(s) not provided."
            raise PydanticCustomError("logger_path_missing", msg)  # noqa: EM101
        return self

    @model_validator(mode="after")
    def check_bounding_not_used_with_data_extractor_zero(self) -> Self:
        """Check that bounding is not used with zero data extractor."""
        if (
            isinstance(self.model.decoder, GraphTransformerDecoderSchema)
            and self.model.decoder.initialise_data_extractor_zero
            and self.model.bounding is not None
        ):
            error = "bounding_conflict_with_data_extractor_zero"
            msg = (
                "Boundings cannot be used with zero initialized weights in decoder. "
                "Set initalise_data_extractor_zero to False."
            )
            raise PydanticCustomError(
                error,
                msg,
            )

    def model_dump(self, by_alias: bool = False) -> dict:
        dumped_model = super().model_dump(by_alias=by_alias)
        return DictConfig(dumped_model)


class UnvalidatedBaseSchema(PydanticBaseModel):
    data: Any
    """Data configuration."""
    dataloader: Any
    """Dataloader configuration."""
    datamodule: Any
    """Datamodule configuration."""
    diagnostics: Any
    """Diagnostics configuration such as logging, plots and metrics."""
    hardware: Any
    """Hardware configuration."""
    graph: Any
    """Graph configuration."""
    model: Any
    """Model configuration."""
    training: Any
    """Training configuration."""
    config_validation: bool = False
    """Flag to disable validation of the configuration"""

    def model_dump(self, by_alias: bool = False) -> dict:
        dumped_model = super().model_dump(by_alias=by_alias)
        return DictConfig(dumped_model)


def convert_to_omegaconf(config: BaseSchema) -> dict:
    config = config.model_dump(by_alias=True)
    return OmegaConf.create(config)


def validate_schema(config: DictConfig) -> BaseSchema:
    try:
        config = BaseSchema(**config)
    except ValidationError as e:
        errors = convert_errors(e, CUSTOM_MESSAGES)
        LOGGER.error(errors)  # noqa: TRY400
        sys.exit(0)
    else:
        return config



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import contextlib
import sys

from mlflow.system_metrics.metrics.base_metrics_monitor import BaseMetricsMonitor

with contextlib.suppress(ImportError):
    import pynvml
with contextlib.suppress(ImportError):
    from pyrsmi import rocml


class GreenGPUMonitor(BaseMetricsMonitor):
    """Class for monitoring Nvidia GPU stats.

    Requires pynvml to be installed.
    Extends default GPUMonitor, to also measure total \
            memory

    """

    def __init__(self):
        if "pynvml" not in sys.modules:
            # Only instantiate if `pynvml` is installed.
            import_error_msg = "`pynvml` is not installed, if you are running on an Nvidia GPU \
                and want to log GPU metrics please run `pip install pynvml`."
            raise ImportError(import_error_msg)
        try:
            # `nvmlInit()` will fail if no GPU is found.
            pynvml.nvmlInit()
        except pynvml.NVMLError as e:
            runtime_error_msg = "Failed to initalize Nvidia GPU monitor: "
            raise RuntimeError(runtime_error_msg) from e

        super().__init__()
        self.num_gpus = pynvml.nvmlDeviceGetCount()
        self.gpu_handles = [pynvml.nvmlDeviceGetHandleByIndex(i) for i in range(self.num_gpus)]

    def collect_metrics(self) -> None:
        # Get GPU metrics.
        for i, handle in enumerate(self.gpu_handles):
            memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
            self._metrics[f"gpu_{i}_memory_usage_percentage"].append(
                round(memory.used / memory.total * 100, 1),
            )
            self._metrics[f"gpu_{i}_memory_usage_megabytes"].append(memory.used / 1e6)

            # Only record total device memory on GPU 0 to prevent spam
            # Unlikely for GPUs on the same node to have different total memory
            if i == 0:
                self._metrics["gpu_memory_total_megabytes"].append(memory.total / 1e6)

            # Monitor PCIe usage
            tx_kilobytes = pynvml.nvmlDeviceGetPcieThroughput(handle, pynvml.NVML_PCIE_UTIL_TX_BYTES)
            rx_kilobytes = pynvml.nvmlDeviceGetPcieThroughput(handle, pynvml.NVML_PCIE_UTIL_RX_BYTES)
            self._metrics[f"gpu_{i}_pcie_tx_megabytes"].append(tx_kilobytes / 1e3)
            self._metrics[f"gpu_{i}_pcie_rx_megabytes"].append(rx_kilobytes / 1e3)

            device_utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
            self._metrics[f"gpu_{i}_utilization_percentage"].append(device_utilization.gpu)

            power_milliwatts = pynvml.nvmlDeviceGetPowerUsage(handle)
            power_capacity_milliwatts = pynvml.nvmlDeviceGetEnforcedPowerLimit(handle)
            self._metrics[f"gpu_{i}_power_usage_watts"].append(power_milliwatts / 1000)
            self._metrics[f"gpu_{i}_power_usage_percentage"].append(
                (power_milliwatts / power_capacity_milliwatts) * 100,
            )

    def aggregate_metrics(self) -> dict[str, int]:
        return {k: round(sum(v) / len(v), 1) for k, v in self._metrics.items()}


class RedGPUMonitor(BaseMetricsMonitor):
    """Class for monitoring AMD GPU stats.

    Requires that pyrsmi is installed
    Logs utilization and memory usage.

    """

    def __init__(self):
        if "pyrsmi" not in sys.modules:
            import_error_msg = "`pyrsmi` is not installed, if you are running on an AMD GPU \
                and want to log GPU metrics please run `pip install pyrsmi`."
            # Only instantiate if `pyrsmi` is installed.
            raise ImportError(import_error_msg)
        try:
            # `rocml.smi_initialize()()` will fail if no GPU is found.
            rocml.smi_initialize()
        except RuntimeError as e:
            runtime_error_msg = "Failed to initalize AMD GPU monitor: "
            raise RuntimeError(runtime_error_msg) from e

        super().__init__()
        self.num_gpus = rocml.smi_get_device_count()

    def collect_metrics(self) -> None:
        # Get GPU metrics.
        for device in range(self.num_gpus):
            memory_used = rocml.smi_get_device_memory_used(device)
            memory_total = rocml.smi_get_device_memory_total(device)
            memory_busy = rocml.smi_get_device_memory_busy(device)
            self._metrics[f"gpu_{device}_memory_usage_percentage"].append(
                round(memory_used / memory_total * 100, 1),
            )
            self._metrics[f"gpu_{device}_memory_usage_megabytes"].append(memory_used / 1e6)

            self._metrics[f"gpu_{device}_memory_busy_percentage"].append(memory_busy)

            # Only record total device memory on GPU 0 to prevent spam
            # Unlikely for GPUs on the same node to have different total memory
            if device == 0:
                self._metrics["gpu_memory_total_megabytes"].append(memory_total / 1e6)

            utilization = rocml.smi_get_device_utilization(device)
            self._metrics[f"gpu_{device}_utilization_percentage"].append(utilization)

    def aggregate_metrics(self) -> dict[str, int]:
        return {k: round(sum(v) / len(v), 1) for k, v in self._metrics.items()}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np
import torch

from anemoi.training.losses.base import FunctionalLoss

LOGGER = logging.getLogger(__name__)


class LogCosh(torch.autograd.Function):
    """LogCosh custom autograd function."""

    @staticmethod
    def forward(ctx, inp: torch.Tensor) -> torch.Tensor:  # noqa: ANN001
        ctx.save_for_backward(inp)
        abs_input = torch.abs(inp)
        return abs_input + torch.nn.functional.softplus(-2 * abs_input) - np.log(2)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:  # noqa: ANN001
        (inp,) = ctx.saved_tensors
        return grad_output * torch.tanh(inp)


class LogCoshLoss(FunctionalLoss):
    """LogCosh loss."""

    name: str = "logcosh"

    def calculate_difference(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Calculate the Log-cosh loss.

        Parameters
        ----------
        pred : torch.Tensor
            Prediction tensor, shape (bs, ensemble, lat*lon, n_outputs)
        target : torch.Tensor
            Target tensor, shape (bs, ensemble, lat*lon, n_outputs)

        Returns
        -------
        torch.Tensor
            Log-cosh loss
        """
        return LogCosh.apply(pred - target)



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Literal

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field


class DataModuleSchema(PydanticBaseModel):
    target_: Literal[
        "anemoi.training.data.datamodule.AnemoiEnsDatasetsDataModule",
        "anemoi.training.data.datamodule.AnemoiDatasetsDataModule",
    ] = Field(..., alias="_target_")



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import os
import shutil
from pathlib import Path

import pytest
from hydra import compose
from hydra import initialize
from omegaconf import OmegaConf


@pytest.fixture(autouse=True)
def set_working_directory() -> None:
    """Automatically set the working directory to the repo root."""
    repo_root = Path(__file__).resolve().parent
    while not (repo_root / ".git").exists() and repo_root != repo_root.parent:
        repo_root = repo_root.parent

    os.chdir(repo_root)


@pytest.fixture
def testing_modifications_with_temp_dir(tmp_path: Path) -> OmegaConf:
    testing_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/testing_modifications.yaml")
    testing_modifications.hardware.paths.output = str(tmp_path)
    return testing_modifications


@pytest.fixture(
    params=[
        ["model=gnn"],
        ["model=graphtransformer"],
    ],
)
def architecture_config(
    request: pytest.FixtureRequest,
    testing_modifications_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, str]:
    overrides = request.param
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_config"):
        template = compose(
            config_name="config",
            overrides=overrides,
        )  # apply architecture overrides to template since they override a default

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = rel_paths[0]

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls[0]


@pytest.fixture
def stretched_config(
    testing_modifications_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, list[str]]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_stretched"):
        template = compose(config_name="stretched")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_stretched.yaml")

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset", "forcing_dataset"])
    dataset, forcing_dataset = rel_paths
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = dataset
    use_case_modifications.hardware.files.forcing_dataset = forcing_dataset

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls


@pytest.fixture
def lam_config(testing_modifications_with_temp_dir: OmegaConf, get_tmp_paths: callable) -> tuple[OmegaConf, str]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_lam"):
        template = compose(config_name="lam")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_lam.yaml")

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset", "forcing_dataset"])
    dataset, forcing_dataset = rel_paths
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = dataset
    use_case_modifications.hardware.files.forcing_dataset = forcing_dataset

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls


@pytest.fixture
def lam_config_with_graph(lam_config: OmegaConf, get_test_data: callable) -> tuple[OmegaConf, list[str]]:
    existing_graph_config = OmegaConf.load(Path.cwd() / "training/src/anemoi/training/config/graph/existing.yaml")
    cfg, urls = lam_config
    cfg.graph = existing_graph_config

    url_graph = cfg.hardware.files["graph"]
    tmp_path_graph = get_test_data(url_graph)
    cfg.hardware.paths.graph = Path(tmp_path_graph).parent
    cfg.hardware.files.graph = Path(tmp_path_graph).name

    return cfg, urls


@pytest.fixture
def ensemble_config(testing_modifications_with_temp_dir: OmegaConf, get_tmp_paths: callable) -> tuple[OmegaConf, str]:
    overrides = ["model=graphtransformer_ens", "graph=multi_scale"]

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_ensemble_crps"):
        template = compose(config_name="ensemble_crps", overrides=overrides)

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_ensemble_crps.yaml")

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = rel_paths[0]

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls[0]


@pytest.fixture
def hierarchical_config(
    testing_modifications_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, list[str]]:
    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_hierarchical"):
        template = compose(config_name="hierarchical")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = rel_paths[0]

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls


@pytest.fixture
def get_tmp_paths(temporary_directory_for_test_data: callable) -> callable:
    def _get_tmp_paths(config: OmegaConf, list_datasets: list[str]) -> tuple[str, list[str], list[str]]:
        tmp_paths = []
        dataset_names = []
        archive_urls = []

        for dataset in list_datasets:
            url_archive = config.hardware.files[dataset] + ".tgz"
            name_dataset = Path(config.hardware.files[dataset]).name
            tmp_path_dataset = temporary_directory_for_test_data(url_archive, archive=True)

            tmp_paths.append(tmp_path_dataset)
            dataset_names.append(name_dataset)
            archive_urls.append(url_archive)

        if len(list_datasets) == 1:
            return tmp_paths[0], dataset_names, archive_urls

        tmp_dir = os.path.commonprefix([tmp_paths[0], tmp_paths[1]])[:-1]  # remove trailing slash
        rel_paths = [Path(Path(path).name) / name for (name, path) in zip(dataset_names, tmp_paths, strict=False)]
        return tmp_dir, rel_paths, archive_urls

    return _get_tmp_paths


@pytest.fixture
def gnn_config(
    testing_modifications_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, str]:

    with initialize(version_base=None, config_path="../../src/anemoi/training/config", job_name="test_config"):
        template = compose(config_name="config")

    use_case_modifications = OmegaConf.load(Path.cwd() / "training/tests/integration/config/test_config.yaml")
    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = rel_paths[0]

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls[0]


@pytest.fixture
def gnn_config_with_checkpoint(gnn_config: OmegaConf, get_test_data: callable) -> OmegaConf:
    cfg, dataset_url = gnn_config
    existing_ckpt = get_test_data(
        "anemoi-integration-tests/training/checkpoints/testing-checkpoint-global-2025-07-31.ckpt",
    )
    checkpoint_dir = Path(cfg.hardware.paths.output + "checkpoint/dummy_id")
    checkpoint_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy(existing_ckpt, checkpoint_dir / "last.ckpt")

    cfg.training.run_id = "dummy_id"
    cfg.training.max_epochs = 3
    return cfg, dataset_url


@pytest.fixture
def interpolator_config(
    testing_modifications_with_temp_dir: OmegaConf,
    get_tmp_paths: callable,
) -> tuple[OmegaConf, str]:
    """Compose a runnable configuration for the temporal-interpolation model.

    It is based on `interpolator.yaml` and only patches paths pointing to the
    sample dataset that the tests download locally.
    """
    # No model override here - the template already sets the dedicated
    # interpolator model + GraphInterpolator Lightning task.
    with initialize(
        version_base=None,
        config_path="../../src/anemoi/training/config",
        job_name="test_interpolator",
    ):
        template = compose(config_name="interpolator")

    use_case_modifications = OmegaConf.load(
        Path.cwd() / "training/tests/integration/config/test_interpolator.yaml",
    )

    tmp_dir, rel_paths, dataset_urls = get_tmp_paths(use_case_modifications, ["dataset"])
    use_case_modifications.hardware.paths.data = tmp_dir
    use_case_modifications.hardware.files.dataset = rel_paths[0]

    cfg = OmegaConf.merge(template, testing_modifications_with_temp_dir, use_case_modifications)
    OmegaConf.resolve(cfg)
    return cfg, dataset_urls[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from dataclasses import dataclass

import datashader as dsh
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.style as mplstyle
import numpy as np
import pandas as pd
from datashader.mpl_ext import dsshow
from matplotlib.collections import LineCollection
from matplotlib.collections import PathCollection
from matplotlib.colors import BoundaryNorm
from matplotlib.colors import Colormap
from matplotlib.colors import Normalize
from matplotlib.colors import TwoSlopeNorm
from matplotlib.figure import Figure
from pyshtools.expand import SHGLQ
from pyshtools.expand import SHExpandGLQ
from scipy.interpolate import griddata
from torch import Tensor
from torch import nn

from anemoi.training.diagnostics.maps import Coastlines
from anemoi.training.diagnostics.maps import EquirectangularProjection
from anemoi.training.utils.variables_metadata import ExtractVariableGroupAndLevel

LOGGER = logging.getLogger(__name__)

continents = Coastlines()
LAYOUT = "tight"


@dataclass
class LatLonData:
    latitudes: np.ndarray
    longitudes: np.ndarray
    data: np.ndarray


def equirectangular_projection(latlons: np.array) -> np.array:
    pc = EquirectangularProjection()
    lat, lon = latlons[:, 0], latlons[:, 1]
    pc_lon, pc_lat = pc(lon, lat)
    return pc_lat, pc_lon


def argsort_variablename_variablelevel(data: list[str], metadata_variables: dict | None = None) -> list[int]:
    """Custom sort key to process the strings.

    Sort parameter names by alpha part, then by numeric part at last
    position (variable level) if available, then by the original string.

    Parameters
    ----------
    data : list[str]
        List of strings to sort.
    metadata_variables : dict, optional
        Dictionary of variable names and indices, by default None

    Returns
    -------
    list[int]
        Sorted indices of the input list.
    """
    extract_variable_group_and_level = ExtractVariableGroupAndLevel(
        {"default": ""},
        metadata_variables,
    )

    def custom_sort_key(index: int) -> tuple:
        s = data[index]  # Access the element by index
        _, alpha_part, numeric_part = extract_variable_group_and_level.get_group_and_level(s)
        if numeric_part is None:
            numeric_part = float("inf")
        return (alpha_part, numeric_part, s)

    # Generate argsort indices
    return sorted(range(len(data)), key=custom_sort_key)


def init_plot_settings() -> None:
    """Initialize matplotlib plot settings."""
    small_font_size = 8
    medium_font_size = 10

    mplstyle.use("fast")
    plt.rcParams["path.simplify_threshold"] = 0.9

    plt.rc("font", size=small_font_size)  # controls default text sizes
    plt.rc("axes", titlesize=small_font_size)  # fontsize of the axes title
    plt.rc("axes", labelsize=medium_font_size)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=small_font_size)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=small_font_size)  # fontsize of the tick labels
    plt.rc("legend", fontsize=small_font_size)  # legend fontsize
    plt.rc("figure", titlesize=small_font_size)  # fontsize of the figure title


def _hide_axes_ticks(ax: plt.Axes) -> None:
    """Hide x/y-axis ticks.

    Parameters
    ----------
    ax : matplotlib.axes
        Axes object handle

    """
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_yticklabels(), visible=False)
    ax.tick_params(axis="both", which="both", length=0)


def plot_loss(
    x: np.ndarray,
    colors: np.ndarray,
    xticks: dict[str, int] | None = None,
    legend_patches: list | None = None,
) -> Figure:
    """Plots data for one multilevel sample.

    Parameters
    ----------
    x : np.ndarray
        Data for Plotting of shape (npred,)
    colors : np.ndarray
        Colors for the bars.
    xticks : dict, optional
        Dictionary of xticks, by default None
    legend_patches : list, optional
        List of legend patches, by default None

    Returns
    -------
    Figure
        The figure object handle.

    """
    # create plot
    # more space for legend
    # TODO(who?): make figsize more flexible depending on the number of bars
    figsize = (8, 3) if legend_patches else (4, 3)
    fig, ax = plt.subplots(1, 1, figsize=figsize, layout=LAYOUT)
    # histogram plot
    ax.bar(np.arange(x.size), x, color=colors, log=1)

    # add xticks and legend if given
    if xticks:
        ax.set_xticks(list(xticks.values()), list(xticks.keys()), rotation=60)
    if legend_patches:
        # legend outside and to the right of the plot
        ax.legend(handles=legend_patches, bbox_to_anchor=(1.01, 1), loc="upper left")

    return fig


def plot_power_spectrum(
    parameters: dict[str, int],
    latlons: np.ndarray,
    x: np.ndarray,
    y_true: np.ndarray,
    y_pred: np.ndarray,
    min_delta: float | None = None,
) -> Figure:
    """Plots power spectrum.

    NB: this can be very slow for large data arrays
    call it as infrequently as possible!

    Parameters
    ----------
    parameters : dict[str, int]
        Dictionary of variable names and indices
    latlons : np.ndarray
        lat/lon coordinates array, shape (lat*lon, 2)
    x : np.ndarray
        Input data of shape (lat*lon, nvar*level)
    y_true : np.ndarray
        Expected data of shape (lat*lon, nvar*level)
    y_pred : np.ndarray
        Predicted data of shape (lat*lon, nvar*level)
    min_delta: float, optional
        Minimum distance between lat/lon points, if None defaulted to 1km

    Returns
    -------
    Figure
        The figure object handle.

    """
    min_delta = min_delta or 0.0003
    n_plots_x, n_plots_y = len(parameters), 1

    figsize = (n_plots_y * 4, n_plots_x * 3)
    fig, ax = plt.subplots(n_plots_x, n_plots_y, figsize=figsize, layout=LAYOUT)
    if n_plots_x == 1:
        ax = [ax]

    pc_lat, pc_lon = equirectangular_projection(latlons)

    pc_lon = np.array(pc_lon)
    pc_lat = np.array(pc_lat)
    # Calculate delta_lat on the projected grid
    delta_lat = abs(np.diff(pc_lat))
    non_zero_delta_lat = delta_lat[delta_lat != 0]
    min_delta_lat = np.min(abs(non_zero_delta_lat))

    if min_delta_lat < min_delta:
        LOGGER.warning(
            "Min. distance between lat/lon points is < specified minimum distance. Defaulting to min_delta=%s.",
            min_delta,
        )
        min_delta_lat = min_delta

    # Define a regular grid for interpolation
    n_pix_lat = int(np.floor(abs(pc_lat.max() - pc_lat.min()) / min_delta_lat))
    n_pix_lon = (n_pix_lat - 1) * 2 + 1  # 2*lmax + 1
    regular_pc_lon = np.linspace(pc_lon.min(), pc_lon.max(), n_pix_lon)
    regular_pc_lat = np.linspace(pc_lat.min(), pc_lat.max(), n_pix_lat)
    grid_pc_lon, grid_pc_lat = np.meshgrid(regular_pc_lon, regular_pc_lat)

    for plot_idx, (variable_idx, (variable_name, output_only)) in enumerate(parameters.items()):
        yt = y_true[..., variable_idx].squeeze()
        yp = y_pred[..., variable_idx].squeeze()

        # check for any nan in yt
        nan_flag = np.isnan(yt).any()

        method = "linear" if nan_flag else "cubic"
        if output_only:
            xt = x[..., variable_idx].squeeze()
            yt_i = griddata((pc_lon, pc_lat), (yt - xt), (grid_pc_lon, grid_pc_lat), method=method, fill_value=0.0)
            yp_i = griddata((pc_lon, pc_lat), (yp - xt), (grid_pc_lon, grid_pc_lat), method=method, fill_value=0.0)
        else:
            yt_i = griddata((pc_lon, pc_lat), yt, (grid_pc_lon, grid_pc_lat), method=method, fill_value=0.0)
            yp_i = griddata((pc_lon, pc_lat), yp, (grid_pc_lon, grid_pc_lat), method=method, fill_value=0.0)

        # Masking NaN values
        if nan_flag:
            mask = np.isnan(yt_i)
            if mask.any():
                yt_i = np.where(mask, 0.0, yt_i)
                yp_i = np.where(mask, 0.0, yp_i)

        amplitude_t = np.array(compute_spectra(yt_i))
        amplitude_p = np.array(compute_spectra(yp_i))

        ax[plot_idx].loglog(
            np.arange(1, amplitude_t.shape[0]),
            amplitude_t[1 : (amplitude_t.shape[0])],
            label="Truth (data)",
        )
        ax[plot_idx].loglog(
            np.arange(1, amplitude_p.shape[0]),
            amplitude_p[1 : (amplitude_p.shape[0])],
            label="Predicted",
        )

        ax[plot_idx].legend()
        ax[plot_idx].set_title(variable_name)

        ax[plot_idx].set_xlabel("$k$")
        ax[plot_idx].set_ylabel("$P(k)$")
        ax[plot_idx].set_aspect("auto", adjustable=None)
    return fig


def compute_spectra(field: np.ndarray) -> np.ndarray:
    """Compute spectral variability of a field by wavenumber.

    Parameters
    ----------
    field : np.ndarray
        lat lon field to calculate the spectra of

    Returns
    -------
    np.ndarray
        spectra of field by wavenumber

    """
    field = np.array(field)

    # compute real and imaginary parts of power spectra of field
    lmax = field.shape[0] - 1  # maximum degree of expansion
    zero_w = SHGLQ(lmax)
    coeffs_field = SHExpandGLQ(field, w=zero_w[1], zero=zero_w[0])

    # Re**2 + Im**2
    coeff_amp = coeffs_field[0, :, :] ** 2 + coeffs_field[1, :, :] ** 2

    # sum over meridional direction
    return np.sum(coeff_amp, axis=0)


def plot_histogram(
    parameters: dict[str, int],
    x: np.ndarray,
    y_true: np.ndarray,
    y_pred: np.ndarray,
    precip_and_related_fields: list | None = None,
) -> Figure:
    """Plots histogram.

    NB: this can be very slow for large data arrays
    call it as infrequently as possible!

    Parameters
    ----------
    parameters : dict[str, int]
        Dictionary of variable names and indices
    x : np.ndarray
        Input data of shape (lat*lon, nvar*level)
    y_true : np.ndarray
        Expected data of shape (lat*lon, nvar*level)
    y_pred : np.ndarray
        Predicted data of shape (lat*lon, nvar*level)
    precip_and_related_fields : list, optional
        List of precipitation-like variables, by default []

    Returns
    -------
    Figure
        The figure object handle.

    """
    precip_and_related_fields = precip_and_related_fields or []

    n_plots_x, n_plots_y = len(parameters), 1

    figsize = (n_plots_y * 4, n_plots_x * 3)
    fig, ax = plt.subplots(n_plots_x, n_plots_y, figsize=figsize, layout=LAYOUT)
    if n_plots_x == 1:
        ax = [ax]

    for plot_idx, (variable_idx, (variable_name, output_only)) in enumerate(parameters.items()):
        yt = y_true[..., variable_idx].squeeze()
        yp = y_pred[..., variable_idx].squeeze()
        # postprocessed outputs so we need to handle possible NaNs

        # Calculate the histogram and handle NaNs
        if output_only:
            # histogram of true increment and predicted increment
            xt = x[..., variable_idx].squeeze() * int(output_only)
            yt_xt = yt - xt
            yp_xt = yp - xt
            # enforce the same binning for both histograms
            bin_min = min(np.nanmin(yt_xt), np.nanmin(yp_xt))
            bin_max = max(np.nanmax(yt_xt), np.nanmax(yp_xt))
            hist_yt, bins_yt = np.histogram(yt_xt[~np.isnan(yt_xt)], bins=100, density=True, range=[bin_min, bin_max])
            hist_yp, bins_yp = np.histogram(yp_xt[~np.isnan(yp_xt)], bins=100, density=True, range=[bin_min, bin_max])
        else:
            # enforce the same binning for both histograms
            bin_min = min(np.nanmin(yt), np.nanmin(yp))
            bin_max = max(np.nanmax(yt), np.nanmax(yp))
            hist_yt, bins_yt = np.histogram(yt[~np.isnan(yt)], bins=100, density=True, range=[bin_min, bin_max])
            hist_yp, bins_yp = np.histogram(yp[~np.isnan(yp)], bins=100, density=True, range=[bin_min, bin_max])

        # Visualization trick for tp
        if variable_name in precip_and_related_fields:
            # in-place multiplication does not work here because variables are different numpy types
            hist_yt = hist_yt * bins_yt[:-1]
            hist_yp = hist_yp * bins_yp[:-1]
        # Plot the modified histogram
        ax[plot_idx].bar(bins_yt[:-1], hist_yt, width=np.diff(bins_yt), color="blue", alpha=0.7, label="Truth (data)")
        ax[plot_idx].bar(bins_yp[:-1], hist_yp, width=np.diff(bins_yp), color="red", alpha=0.7, label="Predicted")

        ax[plot_idx].set_title(variable_name)
        ax[plot_idx].set_xlabel(variable_name)
        ax[plot_idx].set_ylabel("Density")
        ax[plot_idx].legend()
        ax[plot_idx].set_aspect("auto", adjustable=None)

    return fig


def plot_predicted_multilevel_flat_sample(
    parameters: dict[str, int],
    n_plots_per_sample: int,
    latlons: np.ndarray,
    clevels: float,
    x: np.ndarray,
    y_true: np.ndarray,
    y_pred: np.ndarray,
    datashader: bool = False,
    precip_and_related_fields: list | None = None,
    colormaps: dict[str, Colormap] | None = None,
) -> Figure:
    """Plots data for one multilevel latlon-"flat" sample.

    NB: this can be very slow for large data arrays
    call it as infrequently as possible!

    Parameters
    ----------
    parameters : dict[str, int]
        Dictionary of variable names and indices
    n_plots_per_sample : int
        Number of plots per sample
    latlons : np.ndarray
        lat/lon coordinates array, shape (lat*lon, 2)
    clevels : float
        Accumulation levels used for precipitation related plots
    x : np.ndarray
        Input data of shape (lat*lon, nvar*level)
    y_true : np.ndarray
        Expected data of shape (lat*lon, nvar*level)
    y_pred : np.ndarray
        Predicted data of shape (lat*lon, nvar*level)
    datashader: bool, optional
        Scatter plot, by default False
    precip_and_related_fields : list, optional
        List of precipitation-like variables, by default []
    colormaps : dict[str, Colormap], optional
        Dictionary of colormaps, by default None

    Returns
    -------
    Figure
        The figure object handle.

    """
    n_plots_x, n_plots_y = len(parameters), n_plots_per_sample

    figsize = (n_plots_y * 4, n_plots_x * 3)
    fig, ax = plt.subplots(n_plots_x, n_plots_y, figsize=figsize, layout=LAYOUT)

    pc_lat, pc_lon = equirectangular_projection(latlons)
    if colormaps is None:
        colormaps = {}

    for plot_idx, (variable_idx, (variable_name, output_only)) in enumerate(parameters.items()):
        xt = x[..., variable_idx].squeeze() * int(output_only)
        yt = y_true[..., variable_idx].squeeze()
        yp = y_pred[..., variable_idx].squeeze()

        # get the colormap for the variable as defined in config file
        cmap = colormaps.default.get_cmap() if colormaps.get("default") else cm.get_cmap("viridis")
        error_cmap = colormaps.error.get_cmap() if colormaps.get("error") else cm.get_cmap("bwr")
        for key in colormaps:
            if key not in ["default", "error"] and variable_name in colormaps[key].variables:
                cmap = colormaps[key].get_cmap()
                continue
        if n_plots_x > 1:
            plot_flat_sample(
                fig,
                ax[plot_idx, :],
                pc_lon,
                pc_lat,
                xt,
                yt,
                yp,
                variable_name,
                clevels,
                datashader,
                precip_and_related_fields,
                cmap=cmap,
                error_cmap=error_cmap,
            )
        else:
            plot_flat_sample(
                fig,
                ax,
                pc_lon,
                pc_lat,
                xt,
                yt,
                yp,
                variable_name,
                clevels,
                datashader,
                precip_and_related_fields,
                cmap=cmap,
                error_cmap=error_cmap,
            )

    return fig


def plot_flat_sample(
    fig: Figure,
    ax: plt.Axes,
    lon: np.ndarray,
    lat: np.ndarray,
    input_: np.ndarray,
    truth: np.ndarray,
    pred: np.ndarray,
    vname: str,
    clevels: float,
    datashader: bool = False,
    precip_and_related_fields: list | None = None,
    cmap: Colormap | None = None,
    error_cmap: Colormap | None = None,
) -> None:
    """Plot a "flat" 1D sample.

    Data on non-rectangular (reduced Gaussian) grids.

    Parameters
    ----------
    fig : Figure
        Figure object handle
    ax : matplotlib.axes
        Axis object handle
    lon : np.ndarray
        longitude coordinates array, shape (lon,)
    lat : np.ndarray
        latitude coordinates array, shape (lat,)
    input_ : np.ndarray
        Input data of shape (lat*lon,)
    truth : np.ndarray
        Expected data of shape (lat*lon,)
    pred : np.ndarray
        Predicted data of shape (lat*lon,)
    vname : str
        Variable name
    clevels : float
        Accumulation levels used for precipitation related plots
    datashader: bool, optional
        Datashader plott, by default True
    precip_and_related_fields : list, optional
        List of precipitation-like variables, by default []
    cmap : Colormap, optional
        Colormap for the plot
    error_cmap : Colormap, optional
        Colormap for the error plot

    Returns
    -------
    None
    """
    precip_and_related_fields = precip_and_related_fields or []
    if vname in precip_and_related_fields:
        # converting to mm from m
        truth *= 1000.0
        pred *= 1000.0
        if sum(input_) != 0:
            input_ *= 1000.0
    data = [None for _ in range(6)]
    # truth, prediction and prediction error always plotted
    data[1:4] = [truth, pred, truth - pred]
    # default titles for 6 plots
    titles = [
        f"{vname} input",
        f"{vname} target",
        f"{vname} pred",
        f"{vname} pred err",
        f"{vname} increment [pred - input]",
        f"{vname} persist err",
    ]
    # colormaps
    cmaps = [cmap] * 3 + [error_cmap] * 3
    # normalizations for significant colormaps
    norms = [None for _ in range(6)]
    norms[3:6] = [TwoSlopeNorm(vcenter=0.0)] * 3  # center the error colormaps at 0

    if vname in precip_and_related_fields:
        # Defining the actual precipitation accumulation levels in mm
        cummulation_lvls = clevels
        norm = BoundaryNorm(cummulation_lvls, len(cummulation_lvls) + 1)

        norms[1] = norm
        norms[2] = norm

    else:
        combined_data = np.concatenate((input_, truth, pred))
        # For 'errors', only persistence and increments need identical colorbar-limits

        norm = Normalize(vmin=np.nanmin(combined_data), vmax=np.nanmax(combined_data))

        norms[1] = norm
        norms[2] = norm

    if sum(input_) != 0:
        # prognostic fields: plot input and increment as well
        data[0] = input_
        data[4] = pred - input_
        data[5] = truth - input_
        combined_error = np.concatenate(((pred - input_), (truth - input_)))
        # ensure vcenter is between minimum and maximum error
        norm_error = TwoSlopeNorm(
            vmin=min(-0.00001, np.nanmin(combined_error)),
            vcenter=0.0,
            vmax=max(0.00001, np.nanmax(combined_error)),
        )
        norms[0] = norm
        norms[4] = norm_error
        norms[5] = norm_error

    else:
        # diagnostic fields: omit input and increment plots
        ax[0].axis("off")
        ax[4].axis("off")
        ax[5].axis("off")

    for ii in range(6):
        if data[ii] is not None:
            single_plot(
                fig,
                ax[ii],
                lon,
                lat,
                data[ii],
                cmap=cmaps[ii],
                norm=norms[ii],
                title=titles[ii],
                datashader=datashader,
            )


def single_plot(
    fig: Figure,
    ax: plt.axes,
    lon: np.array,
    lat: np.array,
    data: np.array,
    cmap: Colormap | None = None,
    norm: str | None = None,
    title: str | None = None,
    datashader: bool = False,
) -> None:
    """Plot a single lat-lon map.

    Plotting can be made either using datashader plot or Datashader(bin) plots.
    By default it uses Datashader since it is faster and more efficient.

    Parameters
    ----------
    fig : Figure
        Figure object handle
    ax : matplotlib.axes
        Axis object handle
    lon : np.ndarray
        longitude coordinates array, shape (lon,)
    lat : np.ndarray
        latitude coordinates array, shape (lat,)
    data : np.ndarray
        Data to plot
    cmap : Colormap, optional
        Colormap, if None use "viridis"
    norm : str, optional
        Normalization string from matplotlib, by default None
    title : str, optional
        Title for plot, by default None
    datashader: bool, optional
        Scatter plot, by default False

    Returns
    -------
    None
    """
    if cmap is None:
        cmap = "viridis"
    if not datashader:
        psc = ax.scatter(
            lon,
            lat,
            c=data,
            cmap=cmap,
            s=1,
            alpha=1.0,
            norm=norm,
            rasterized=False,
        )
    else:
        df = pd.DataFrame({"val": data, "x": lon, "y": lat})
        # Adjust binning to match the resolution of the data
        lower_limit = 25
        upper_limit = 500
        n_pixels = max(min(int(np.floor(data.shape[0] * 0.004)), upper_limit), lower_limit)
        psc = dsshow(
            df,
            dsh.Point("x", "y"),
            dsh.mean("val"),
            cmap=cmap,
            plot_width=n_pixels,
            plot_height=n_pixels,
            norm=norm,
            aspect="auto",
            ax=ax,
        )

    xmin, xmax = max(lon.min(), -np.pi), min(lon.max(), np.pi)
    ymin, ymax = max(lat.min(), -np.pi / 2), min(lat.max(), np.pi / 2)
    ax.set_xlim((xmin - 0.1, xmax + 0.1))
    ax.set_ylim((ymin - 0.1, ymax + 0.1))

    continents.plot_continents(ax)

    if title is not None:
        ax.set_title(title)

    ax.set_aspect("auto", adjustable=None)
    _hide_axes_ticks(ax)
    fig.colorbar(psc, ax=ax)


def get_scatter_frame(
    ax: plt.Axes,
    data: np.ndarray,
    latlons: np.ndarray,
    cmap: str = "viridis",
    vmin: int | None = None,
    vmax: int | None = None,
) -> [plt.Axes, PathCollection]:
    """Create a scatter plot for a single frame of an animation."""
    pc_lat, pc_lon = equirectangular_projection(latlons)

    scatter_frame = ax.scatter(
        pc_lon,
        pc_lat,
        c=data,
        cmap=cmap,
        s=5,
        alpha=1.0,
        rasterized=True,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_xlim((-np.pi, np.pi))
    ax.set_ylim((-np.pi / 2, np.pi / 2))
    continents.plot_continents(ax)
    ax.set_aspect("auto", adjustable=None)
    _hide_axes_ticks(ax)
    return ax, scatter_frame


def edge_plot(
    fig: Figure,
    ax: plt.Axes,
    src_coords: np.ndarray,
    dst_coords: np.ndarray,
    data: np.ndarray,
    cmap: str = "coolwarm",
    title: str | None = None,
) -> None:
    """Lat-lon line plot.

    Parameters
    ----------
    fig : Figure
        Figure object handle
    ax : matplotlib.axes
        Axis object handle
    src_coords : np.ndarray of shape (num_edges, 2)
        Source latitudes and longitudes.
    dst_coords : np.ndarray of shape (num_edges, 2)
        Destination latitudes and longitudes.
    data : np.ndarray of shape (num_edges, 1)
        Data to plot
    cmap : str, optional
        Colormap string from matplotlib, by default "viridis".
    title : str, optional
        Title for plot, by default None
    """
    edge_lines = np.stack([src_coords, dst_coords], axis=1)
    lc = LineCollection(edge_lines, cmap=cmap, linewidths=1)
    lc.set_array(data)

    psc = ax.add_collection(lc)

    xmin, xmax = edge_lines[:, 0, 0].min(), edge_lines[:, 0, 0].max()
    ymin, ymax = edge_lines[:, 1, 1].min(), edge_lines[:, 1, 1].max()
    ax.set_xlim((xmin - 0.1, xmax + 0.1))
    ax.set_ylim((ymin - 0.1, ymax + 0.1))

    continents.plot_continents(ax)

    if title is not None:
        ax.set_title(title)

    ax.set_aspect("auto", adjustable=None)
    _hide_axes_ticks(ax)
    fig.colorbar(psc, ax=ax)


def plot_graph_node_features(
    model: nn.Module,
    trainable_tensors: dict[str, Tensor],
    datashader: bool = False,
) -> Figure:
    """Plot trainable graph node features.

    Parameters
    ----------
    model: AneomiModelEncProcDec
        Model object
    trainable_tensors: dict[str, torch.Tensor]
        Node trainable tensors
    datashader: bool, optional
        Scatter plot, by default False

    Returns
    -------
    Figure
        Figure object handle
    """
    nrows = len(trainable_tensors)
    ncols = max(tt.shape[1] for tt in trainable_tensors.values())

    figsize = (ncols * 4, nrows * 3)
    fig, ax = plt.subplots(nrows, ncols, figsize=figsize, layout=LAYOUT)

    for row, (mesh, trainable_tensor) in enumerate(trainable_tensors.items()):
        latlons = model.node_attributes.get_coordinates(mesh).cpu().numpy()
        node_features = trainable_tensor.cpu().detach().numpy()

        lat, lon = latlons[:, 0], latlons[:, 1]

        for i in range(ncols):
            ax_ = ax[row, i] if ncols > 1 else ax[row]
            single_plot(
                fig,
                ax_,
                lon=lon,
                lat=lat,
                data=node_features[..., i],
                title=f"{mesh} trainable feature #{i + 1}",
                datashader=datashader,
            )

    return fig


def plot_graph_edge_features(
    model: nn.Module,
    trainable_modules: dict[tuple[str, str], Tensor],
    q_extreme_limit: float = 0.05,
) -> Figure:
    """Plot trainable graph edge features.

    Parameters
    ----------
    model: AneomiModelEncProcDec
        Model object
    trainable_modules: dict[tuple[str, str], torch.Tensor]
        Edge trainable tensors.
    q_extreme_limit : float, optional
        Plot top & bottom quantile of edges trainable values, by default 0.05 (5%).

    Returns
    -------
    Figure
        Figure object handle
    """
    nrows = len(trainable_modules)
    ncols = max(tt.trainable.trainable.shape[1] for tt in trainable_modules.values())
    figsize = (ncols * 4, nrows * 3)
    fig, ax = plt.subplots(nrows, ncols, figsize=figsize, layout=LAYOUT)

    for row, ((src, dst), graph_mapper) in enumerate(trainable_modules.items()):
        src_coords = model.node_attributes.get_coordinates(src).cpu().numpy()
        dst_coords = model.node_attributes.get_coordinates(dst).cpu().numpy()
        edge_index = graph_mapper.edge_index_base.cpu().numpy()
        edge_features = graph_mapper.trainable.trainable.cpu().detach().numpy()

        for i in range(ncols):
            ax_ = ax[row, i] if ncols > 1 else ax[row]
            feature = edge_features[..., i]

            # Get mask of feature values over top and bottom percentiles
            top_perc = np.quantile(feature, 1 - q_extreme_limit)
            bottom_perc = np.quantile(feature, q_extreme_limit)

            mask = (feature >= top_perc) | (feature <= bottom_perc)

            edge_plot(
                fig,
                ax_,
                src_coords[edge_index[0, mask]][:, ::-1],
                dst_coords[edge_index[1, mask]][:, ::-1],
                feature[mask],
                title=f"{src} -> {dst} trainable feature #{i + 1}",
            )

    return fig



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from anemoi.training.train.train import AnemoiTrainer


@pytest.fixture(scope="session")
def tmp_checkpoint_factory(tmp_path_factory: pytest.TempPathFactory) -> (Path, Path):
    def _create_checkpoint(
        rid: str | None = None,
        ckpt_path_name: str = "mock_checkpoints",
        ckpt_file_name: str = "last.ckpt",
        skip_creation: bool = False,
    ) -> (Path, Path):
        base = tmp_path_factory.mktemp(ckpt_path_name)
        checkpoint_dir = base / rid if rid else base
        checkpoint_dir.mkdir(parents=True, exist_ok=True)
        ckpt_file = checkpoint_dir / ckpt_file_name
        if not skip_creation:
            ckpt_file.write_text("fake checkpoint")
        return ckpt_file, base  # full path to the .ckpt

    return _create_checkpoint


def build_mock_config(
    *,
    run_id: str | None = None,
    fork_run_id: str | None = None,
    warm_start: str | None = None,
    checkpoints_path: Path | None = None,
    warm_start_path: Path | None = None,
) -> MagicMock:
    config = MagicMock()
    config.config_validation = False
    config.training.run_id = run_id
    config.training.fork_run_id = fork_run_id
    config.training.load_weights_only = False
    config.training.transfer_learning = False
    config.hardware.files.warm_start = warm_start
    config.hardware.paths.checkpoints = checkpoints_path
    config.hardware.paths.warm_start = warm_start_path
    config.diagnostics.log.mlflow.enabled = False
    return config


@pytest.fixture
def trainer_factory() -> AnemoiTrainer:
    def _make_trainer(mock_config: MagicMock) -> AnemoiTrainer:
        with patch("anemoi.training.train.train.OmegaConf.to_object", return_value=mock_config), patch(
            "anemoi.training.train.train.DictConfig",
            return_value=mock_config,
        ), patch("anemoi.training.train.train.UnvalidatedBaseSchema", return_value=mock_config), patch(
            "anemoi.training.train.train.LOGGER",
        ), patch(
            "anemoi.training.train.train.AnemoiTrainer._check_dry_run",
        ), patch(
            "anemoi.training.train.train.AnemoiTrainer._log_information",
        ), patch(
            "pathlib.Path.exists",
            return_value=True,
        ):
            return AnemoiTrainer(config=mock_config)

    return _make_trainer


def test_restart_run_id(trainer_factory: AnemoiTrainer, tmp_checkpoint_factory: pytest.TempPathFactory) -> None:
    run_id = "run-id-123"
    expected_path, checkpoints_path = tmp_checkpoint_factory(rid=run_id, ckpt_path_name="mock_checkpoints")

    config = build_mock_config(
        run_id=run_id,
        checkpoints_path=checkpoints_path,
        warm_start_path=None,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id
    assert trainer.last_checkpoint == expected_path


def test_restart_fork_run_id(trainer_factory: AnemoiTrainer, tmp_checkpoint_factory: pytest.TempPathFactory) -> None:
    fork_run_id = "fork-id-456"
    # Forking assumes that the checkpoint is still stored in the same root path as where we
    # will be writing new checkpoints in the training run
    expected_path, checkpoints_path = tmp_checkpoint_factory(rid=fork_run_id, ckpt_path_name="mock_checkpoints")

    config = build_mock_config(
        fork_run_id=fork_run_id,
        checkpoints_path=checkpoints_path,
        warm_start_path=None,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id != fork_run_id
    assert trainer.last_checkpoint == expected_path


def test_restart_warm_start_path(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test by assuming warm start is unlinked to run or fork run ids."""
    warm_start = "checkpoint_10.ckpt"
    warm_start_fork_run_id = "fork-id-446"
    warm_start_path = "mock-pretrained-checkpoints"
    expected_path, warm_start_path = tmp_checkpoint_factory(
        rid=warm_start_fork_run_id,
        ckpt_file_name=warm_start,
        ckpt_path_name=warm_start_path,
    )
    _, checkpoints_path = tmp_checkpoint_factory()  # path where writing the checkpoints it's different
    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start_path=warm_start_path / Path(warm_start_fork_run_id),  # needs to pass full path
        warm_start=warm_start,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id != warm_start_fork_run_id
    assert trainer.last_checkpoint == expected_path


def test_restart_warm_start_fork_run_id(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test by assuming warm start is linked to fork run id."""
    warm_start = "checkpoint_10.ckpt"
    fork_run_id = "fork-id-446"
    warm_start_path = "mock-pretrained-checkpoints"
    expected_path, warm_start_path = tmp_checkpoint_factory(
        rid=fork_run_id,
        ckpt_file_name=warm_start,
        ckpt_path_name=warm_start_path,
    )
    _, checkpoints_path = tmp_checkpoint_factory()  # path where writing the checkpoints it's different

    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start_path=warm_start_path / Path(fork_run_id),  #
        fork_run_id=fork_run_id,
        warm_start=warm_start,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id != fork_run_id
    assert trainer.last_checkpoint == expected_path


def test_restart_warm_start_run_id(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test by assuming warm start is linked to run id.

    This resembles the case where we want to resume run
    using a checkpoint different from last.ckpt.
    """
    warm_start = "checkpoint_10.ckpt"
    run_id = "id-222"
    warm_start_path = "mock-checkpoints"
    expected_path, warm_start_path = tmp_checkpoint_factory(
        rid=run_id,
        ckpt_file_name=warm_start,
        ckpt_path_name=warm_start_path,
    )
    _, checkpoints_path = tmp_checkpoint_factory(
        ckpt_path_name=warm_start_path,
    )  # path where writing the checkpoints it's different

    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start_path=warm_start_path / Path(run_id),  #
        warm_start=warm_start,
        run_id=run_id,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id
    assert trainer.last_checkpoint == expected_path


def test_restart_warm_start_misconfiguration_path(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test to assert misconfiguration is missing path for warm_start."""
    warm_start = "checkpoint_10.ckpt"
    run_id = "id-222"
    _, checkpoints_path = tmp_checkpoint_factory()

    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start=warm_start,
        run_id=run_id,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id

    # This should raise because the warm start file is not in the expected location
    with pytest.raises(
        AssertionError,
        match=r"Please configure config.hardware.paths.warm_start correctly, found: None",
    ):
        _ = trainer.last_checkpoint


def test_restart_warm_start_misconfiguration_filename(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test to assert misconfiguration is missing file for warm_start."""
    run_id = "id-222"
    warm_start_path = "mock-checkpoints"
    _, warm_start_path = tmp_checkpoint_factory(
        rid=run_id,
        ckpt_path_name=warm_start_path,
    )
    _, checkpoints_path = tmp_checkpoint_factory(
        ckpt_path_name=warm_start_path,
    )  # path where writing the checkpoints it's different

    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start_path=warm_start_path / Path(run_id),  #
        run_id=run_id,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id
    # This should raise because the warm start file is not in the expected location
    with pytest.raises(
        AssertionError,
        match=r"Please configure config.hardware.files.warm_start correctly, found: None",
    ):
        _ = trainer.last_checkpoint


def test_warm_start_file_not_found(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test to assert file not found for warm_start."""
    warm_start = "checkpoint_10.ckpt"
    run_id = "id-222"
    warm_start_path = "mock-checkpoints"
    _, warm_start_path = tmp_checkpoint_factory(
        rid=run_id,
        ckpt_file_name=warm_start,
        ckpt_path_name=warm_start_path,
        skip_creation=True,
    )
    _, checkpoints_path = tmp_checkpoint_factory(
        ckpt_path_name=warm_start_path,
    )  # path where writing the checkpoints it's different

    config = build_mock_config(
        checkpoints_path=checkpoints_path,
        warm_start_path=warm_start_path / Path(run_id),  #
        warm_start=warm_start,
        run_id=run_id,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id
    with pytest.raises(AssertionError, match=r"Warm start checkpoint not found"):
        _ = trainer.last_checkpoint


def test_restart_run_id_file_not_found(
    trainer_factory: AnemoiTrainer,
    tmp_checkpoint_factory: pytest.TempPathFactory,
) -> None:
    """Test to assert file not found for resuming."""
    run_id = "run-id-123"
    _, checkpoints_path = tmp_checkpoint_factory(rid=run_id, ckpt_path_name="mock_checkpoints", skip_creation=True)

    config = build_mock_config(
        run_id=run_id,
        checkpoints_path=checkpoints_path,
        warm_start_path=None,
    )

    trainer = trainer_factory(config)

    assert trainer.start_from_checkpoint is True
    assert trainer.run_id == run_id
    with pytest.raises(RuntimeError, match=r"Could not find last checkpoint"):
        _ = trainer.last_checkpoint



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import numpy as np

from anemoi.models.interface import AnemoiModelInterface
from anemoi.training.utils.enums import TensorDim

LOGGER = logging.getLogger(__name__)
SCALER_DTYPE = tuple[tuple[int], np.ndarray]


class BaseScaler(ABC):
    """Base class for all loss scalers."""

    scale_dims: tuple[TensorDim] = None

    def __init__(self, norm: str | None = None) -> None:
        """Initialise BaseScaler.

        Parameters
        ----------
        norm : str, optional
            Type of normalization to apply. Options are None, unit-sum, unit-mean and l1.
        """
        self.norm = norm
        assert norm in [
            None,
            "unit-sum",
            "l1",
            "unit-mean",
        ], f"{self.__class__.__name__}.norm must be one of: None, unit-sum, l1, unit-mean"
        assert self.scale_dims is not None, f"Class {self.__class__.__name__} must define 'scale_dims'"
        if isinstance(self.scale_dims, TensorDim):
            self.scale_dims = (self.scale_dims,)

    @abstractmethod
    def get_scaling_values(self, **kwargs) -> np.ndarray:
        """Abstract method to get loss scaling."""
        ...

    def normalise(self, values: np.ndarray) -> np.ndarray:
        """Normalise the scaler values."""
        if self.norm is None:
            return values

        if self.norm.lower() in ["l1", "unit-sum"]:
            return values / np.sum(values)

        if self.norm.lower() == "unit-mean":
            return values / np.mean(values)

        error_msg = f"{self.norm} must be one of: None, unit-sum, l1, unit-mean."
        raise ValueError(error_msg)

    def get_scaling(self) -> SCALER_DTYPE:
        """Get scaler.

        Returns
        -------
        scale_dims : tuple[int]
            Dimensions over which the scalers are applied.
        scaler_values : np.ndarray
            Scaler values
        """
        scaler_values = self.get_scaling_values()
        scaler_values = self.normalise(scaler_values)
        scale_dims = tuple(x.value for x in self.scale_dims)
        return scale_dims, scaler_values


class BaseDelayedScaler(BaseScaler):
    """Base class for delayed Scalers.

    The delayed scalers are only initialised when creating all the scalers, but its value is
    computed during the first iteration of the training loop. This delayed scalers are suitable
    for scalers requiring information from the `model.pre_processors`.
    """

    @abstractmethod
    def get_delayed_scaling_values(self, **kwargs) -> np.ndarray: ...

    def get_delayed_scaling(self, model: AnemoiModelInterface) -> SCALER_DTYPE:
        scaler_values = self.get_delayed_scaling_values(model)
        scaler_values = self.normalise(scaler_values)
        scale_dims = tuple(x.value for x in self.scale_dims)
        return scale_dims, scaler_values



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from pathlib import Path
from typing import Any
from typing import Literal

from omegaconf import DictConfig
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import PositiveInt
from pydantic import RootModel
from pydantic import computed_field

from anemoi.utils.dates import frequency_to_timedelta
from anemoi.utils.schemas import BaseModel


class Frequency(RootModel):
    root: Any

    @computed_field
    def as_timedelta(self) -> datetime.timedelta:
        return frequency_to_timedelta(self.root)

    @computed_field
    def as_string(self) -> str:
        delta = self.as_timedelta

        if delta.days > 0 and delta.seconds == 0:
            return f"{delta.days}d"

        if delta.days == 0 and delta.seconds >= 3600 and delta.seconds % 3600 == 0:
            return f"{delta.seconds // 3600}h"

        if delta.days == 0 and delta.seconds >= 60 and delta.seconds % 60 == 0:
            return f"{delta.seconds // 60}m"

        if delta.days == 0 and delta.seconds < 60:
            return f"{delta.seconds}s"

        return str(delta)

    @computed_field
    def as_seconds(self) -> int:
        return int(self.as_timedelta.total_seconds())


class DatasetSchema(PydanticBaseModel):
    """Dataset configuration schema."""

    dataset: str | dict | Path | list[dict] | None = None
    "Dataset, see anemoi-datasets"
    start: str | int | None = Field(default=None)
    "Starting datetime for sample of the dataset."
    end: str | int | None = Field(default=None)
    "Ending datetime [inclusive] for sample of the dataset."
    frequency: Frequency
    "Temporal resolution, frequency must be >= to dataset frequency."
    drop: list | None = Field(default=None)
    "List of variables to drop from dataset"


class LoaderSet(BaseModel):
    training: PositiveInt | None = Field(example=None)
    "Value for training dataset"
    validation: PositiveInt | None = Field(example=None)
    "Value for validation dataset"
    test: PositiveInt | None = Field(example=None)
    "Value for test dataset"


class FullGridIndicesSchema(BaseModel):
    target_: Literal["anemoi.training.data.grid_indices.FullGrid"] = Field(
        "anemoi.training.data.grid_indices.FullGrid",
        alias="_target_",
    )
    "Grid indices for full grid class implementation from anemoi.training.data.grid_indices."
    nodes_name: str = Field(examples=["data"])
    "Name of the grid nodes."


class MaskedGridIndicesSchema(BaseModel):
    target_: Literal["anemoi.training.data.grid_indices.MaskedGrid"] = Field(
        "anemoi.training.data.grid_indices.MaskedGrid",
        alias="_target_",
    )
    "Grid indices for masked grid class implementation from anemoi.training.data.grid_indices."
    nodes_name: str = Field(examples=["data"])
    "Name of the grid nodes."
    node_attribute_name: str = Field(examples=["indices_connected_nodes"])
    "Name of the nodes graph attribute used for masking."


class DataLoaderSchema(PydanticBaseModel):

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    prefetch_factor: int = Field(example=2, ge=0)
    "Number of batches loaded in advance by each worker."
    pin_memory: bool = Field(example=True)
    "If True, the data loader will copy Tensors into device/CUDA pinned memory before returning them."
    num_workers: LoaderSet
    "Number of process per-GPU for batch distribution."
    batch_size: LoaderSet
    "Per-GPU batch size."
    limit_batches: LoaderSet = Field(example=None)
    "Limit number of batches to run. Default value null, will run on all the batches."
    training: DatasetSchema | DictConfig
    "Training DatasetSchema."
    validation: DatasetSchema | DictConfig
    "Validation DatasetSchema."
    test: DatasetSchema | DictConfig
    "Test DatasetSchema."
    validation_rollout: PositiveInt = Field(example=1)
    "Number of rollouts to use for validation, must be equal or greater than rollout expected by callbacks."
    # TODO(Helen): Check that this equal or greater than the number of rollouts expected by callbacks ???
    read_group_size: PositiveInt = Field(example=None)
    "Number of GPUs per reader group. Defaults to number of GPUs (see BaseSchema validators)."
    grid_indices: FullGridIndicesSchema | MaskedGridIndicesSchema
    "Grid indice schema."



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch

from anemoi.training.losses.scaler_tensor import ScaleTensor


def test_scale_contains() -> None:
    scale = ScaleTensor(test=(0, 2))
    assert "test" in scale


def test_scale_contains_indexing() -> None:
    scale = ScaleTensor(test=(0, 2))
    assert 0 in scale


def test_scale_tuple_contains_indexing() -> None:
    scale = ScaleTensor(test=((0, 1), 2))
    assert (0, 1) in scale


def test_scale_tuple_not_contains_indexing() -> None:
    scale = ScaleTensor(test=(0, 2))
    assert (0, 1) not in scale


def test_scale_contains_subset_indexing() -> None:
    scale = ScaleTensor(test=(0, 2), wow=(0, 2))
    assert "test" in scale
    scale = scale.subset("wow")
    assert "wow" in scale
    assert "test" not in scale


def test_scale_contains_subset_by_dim_indexing() -> None:
    scale = ScaleTensor(test=(0, 2), wow=(1, 2))
    assert "test" in scale
    scale = scale.subset_by_dim(1)
    assert "wow" in scale
    assert "test" not in scale


def test_scale_resolve() -> None:
    scale = ScaleTensor(test=(-1, 2))
    scale = scale.resolve(4)
    assert 3 in scale
    assert -1 not in scale


def test_add_existing_scaler() -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])))
    with pytest.raises(ValueError, match=r".*already exists.*"):
        scale.add_scaler(0, torch.tensor(3.0), name="test")


def test_update_scaler() -> None:
    scale = ScaleTensor(test=(0, torch.ones(2)))
    scale.update_scaler("test", torch.tensor([3.0]))
    torch.testing.assert_close(scale.tensors["test"][1], torch.tensor([3.0]))


def test_update_missing_scaler() -> None:
    scale = ScaleTensor(test=(0, torch.ones(2)))
    with pytest.raises(ValueError, match=r".*not found in scalers.*"):
        scale.update_scaler("test_missing", torch.tensor([3.0]))
    assert "test" in scale
    assert (0,) in scale


def test_update_scaler_wrong_dim() -> None:
    scale = ScaleTensor(test=(0, torch.ones((2, 3))))
    with pytest.raises(ValueError, match=r".*does not match shape of saved scaler.*"):
        scale.update_scaler("test", torch.ones((2, 2)))
    assert "test" in scale
    assert 0 in scale


def test_update_scaler_wrong_dim_allow_override() -> None:
    scale = ScaleTensor(test=(0, torch.ones((2, 3))))
    assert scale.update_scaler("test", torch.ones((2, 2)), override=True) is None


@pytest.mark.parametrize(
    ("scalers", "input_tensor", "output"),
    [
        ([[0, torch.Tensor([2])]], torch.tensor([1.0, 2.0, 3.0]), torch.tensor([2.0, 4.0, 6.0])),
        ([[0, torch.Tensor([0.5])]], torch.tensor([10.0, 20.0, 30.0]), torch.tensor([5.0, 10.0, 15.0])),
        ([[-1, torch.Tensor([0.5])]], torch.tensor([10.0, 20.0, 30.0]), torch.tensor([5.0, 10.0, 15.0])),
        ([[0, torch.Tensor([0])]], torch.tensor([5.0, 10.0, 15.0]), torch.tensor([0.0, 0.0, 0.0])),
        (
            [[0, torch.Tensor([0.5])], [-1, torch.Tensor([3])]],
            torch.tensor([10.0, 20.0, 30.0]),
            torch.tensor([15.0, 30.0, 45.0]),
        ),
        (
            [[0, torch.Tensor([0.5])], [0, torch.Tensor([3])]],
            torch.tensor([10.0, 20.0, 30.0]),
            torch.tensor([15.0, 30.0, 45.0]),
        ),
    ],
)
def test_scale_tensor_one_dim(
    scalers: list[list[int, torch.Tensor]],
    input_tensor: torch.Tensor,
    output: torch.Tensor,
) -> None:

    scale = ScaleTensor()
    for scaler in scalers:
        scale.add_scaler(*scaler)

    torch.testing.assert_close(scale.scale(input_tensor), output)
    torch.testing.assert_close(scale.scale_iteratively(input_tensor), output)


def test_invalid_dim_sizes() -> None:
    scaler = ScaleTensor()
    scaler.add_scaler(0, torch.ones((5,)))

    with pytest.raises(ValueError, match=r"Validating tensor 'invalid' raised an error."):
        scaler.add_scaler(0, torch.ones((15,)), name="invalid")


def test_invalid_dim_sizes_negative_indexing() -> None:
    scaler = ScaleTensor()
    scaler.add_scaler(0, torch.ones((5,)))
    scaler.add_scaler(-1, torch.ones((15,)), name="invalid")

    with pytest.raises(ValueError, match=r"Validating tensor 'invalid' raised an error."):
        scaler.resolve(1)


def test_valid_dim_sizes_negative_indexing() -> None:
    scaler = ScaleTensor()
    scaler.add_scaler(0, torch.ones((5,)))
    scaler.add_scaler(-1, torch.ones((15,)))

    scaler.resolve(2)


@pytest.mark.parametrize(
    ("scalers", "input_tensor", "output"),
    [
        ([[0, 2.0]], torch.ones([4, 6]), torch.ones([4, 6]) * 2),
        ([[0, [[1.0, 1.0], [1.0, 2.0]]]], torch.ones((2, 2)), [[1, 1], [1, 2]]),
        ([[(0, 1), [[1.0, 1.0], [1.0, 2.0]]]], torch.ones((2, 2)), [[1, 1], [1, 2]]),
        ([[(1, 0), [[1.0, 1.0], [1.0, 2.0]]]], torch.ones((2, 2)), [[1, 1], [1, 2]]),
        ([[(0, 1), [[1.0, 2.0], [1.0, 1.0]]]], torch.ones((2, 2)), [[1, 2], [1, 1]]),
        ([[(1, 0), [[1.0, 2.0], [1.0, 1.0]]]], torch.ones((2, 2)), [[1, 1], [2, 1]]),
    ],
)
def test_scale_tensor_two_dim(
    scalers: list[list[int, torch.Tensor]],
    input_tensor: torch.Tensor,
    output: torch.Tensor,
) -> None:

    scale = ScaleTensor()
    for scaler in scalers:
        scale.add_scaler(*scaler)

    if not isinstance(input_tensor, torch.Tensor):
        input_tensor = torch.tensor(input_tensor, dtype=torch.float32)
    if not isinstance(output, torch.Tensor):
        output = torch.tensor(output, dtype=torch.float32)

    torch.testing.assert_close(scale.scale(input_tensor), output)
    torch.testing.assert_close(scale.scale_iteratively(input_tensor), output)


@pytest.mark.parametrize("subset_id", ["test", 0])
def test_scaler_subset(subset_id) -> None:  # noqa: ANN001
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    subset = scale.subset(subset_id)
    assert "test" in subset
    assert "wow" not in subset
    assert 0 in subset
    assert 1 not in subset


@pytest.mark.parametrize("without_id", ["test", 0])
def test_scaler_subset_without(without_id: int) -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    subset = scale.without(without_id)
    assert "test" not in subset
    assert "wow" in subset
    assert 1 in subset


@pytest.mark.parametrize("subset_id", [0, 1])
def test_scaler_subset_by_dim(subset_id: int) -> None:
    scale = ScaleTensor(test=(-2, torch.tensor([2.0])), wow=(-1, torch.tensor([3.0])))
    scale = scale.resolve(2)
    subset1 = scale.subset(subset_id)
    assert subset_id in subset1


@pytest.mark.parametrize("without_id", ["test"])
def test_scaler_subset_without_overlap(without_id) -> None:  # noqa: ANN001
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(0, torch.tensor([3.0])))
    subset = scale.without(without_id)
    assert "test" not in subset
    assert "wow" in subset
    assert 0 in subset


def test_scaler_remove_str() -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    subset = scale.remove_scaler("wow")
    assert "test" in subset
    assert "wow" not in subset
    assert 0 in subset


def test_scaler_remove_int() -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    subset = scale.remove_scaler(1)
    assert "test" in subset
    assert "wow" not in subset
    assert 0 in subset
    assert 1 not in subset


def test_scaler_freeze_str() -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    with scale.freeze_state():
        subset = scale.remove_scaler("wow")
        assert "test" in subset
        assert "wow" not in subset
        assert 0 in subset
        assert 1 not in subset

    assert "wow" in subset
    assert 1 in subset


def test_scaler_freeze_int() -> None:
    scale = ScaleTensor(test=(0, torch.tensor([2.0])), wow=(1, torch.tensor([3.0])))
    with scale.freeze_state():
        subset = scale.remove_scaler(1)
        assert "test" in subset
        assert "wow" not in subset
        assert 0 in subset
        assert 1 not in subset

    assert "wow" in subset
    assert 1 in subset



from unittest.mock import Mock

import pytest

from anemoi.training.diagnostics.mlflow.logger import AnemoiMLflowLogger


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> str:
    # returns a session-scoped temporary directory
    return str(tmp_path_factory.mktemp("mlruns"))


@pytest.fixture
def default_logger(tmp_path: str) -> AnemoiMLflowLogger:
    logger = AnemoiMLflowLogger(
        experiment_name="test_experiment",
        run_name="test_run",
        offline=True,
        tracking_uri=None,
        authentication=False,
        save_dir=tmp_path,
    )
    mock_client = Mock()
    logger._mlflow_client = mock_client

    return logger


def test_mlflowlogger_params_limit(default_logger: AnemoiMLflowLogger) -> None:

    default_logger._max_params_length = 3
    params = {"lr": 0.001, "path": "era5", "anemoi.version": 1.5, "bounding": True}
    # # Expect an exception when logging too many hyperparameters
    with pytest.raises(ValueError, match=r"Too many params:"):
        default_logger.log_hyperparams(params)


def test_mlflowlogger_metric_deduplication(default_logger: AnemoiMLflowLogger) -> None:

    default_logger.log_metrics({"foo": 1.0}, step=5)
    default_logger.log_metrics({"foo": 1.0}, step=5)  # duplicate
    # Only the first metric should be logged
    assert len(default_logger._logged_metrics) == 1
    assert next(iter(default_logger._logged_metrics))[0] == "foo"  # key
    assert next(iter(default_logger._logged_metrics))[1] == 5  # step



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import warnings
from datetime import datetime
from datetime import timezone
from functools import cached_property
from pathlib import Path

import hydra
import pandas as pd
import pytorch_lightning as pl
from omegaconf import DictConfig
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities import rank_zero_only
from rich.console import Console

from anemoi.training.data.datamodule import AnemoiDatasetsDataModule
from anemoi.training.diagnostics.profilers import BenchmarkProfiler
from anemoi.training.diagnostics.profilers import ProfilerProgressBar
from anemoi.training.train.train import AnemoiTrainer

LOGGER = logging.getLogger(__name__)
console = Console(record=True, width=200)


class AnemoiProfiler(AnemoiTrainer):
    """Profiling for Anemoi."""

    def __init__(self, config: DictConfig) -> None:
        super().__init__(config)

    def print_report(self, title: str, dataframe: pd.DataFrame, color: str = "white", emoji: str = "") -> None:
        if title == "Model Summary":
            console.print(f"[bold {color}]{title}[/bold {color}]", f":{emoji}:")
            console.print(dataframe, end="\n\n")
        else:
            console.print(f"[bold {color}]{title}[/bold {color}]", f":{emoji}:")
            console.print(dataframe.to_markdown(headers="keys", tablefmt="psql"), end="\n\n")

    @staticmethod
    def print_title() -> None:
        console.print("[bold magenta] Benchmark Profiler Summary [/bold magenta]!", ":book:")

    @staticmethod
    def print_metadata() -> None:
        console.print(f"[bold blue] SLURM NODE(s) {os.getenv('SLURM_JOB_NODELIST', '')} [/bold blue]!")
        console.print(f"[bold blue] SLURM JOB ID {os.getenv('SLURM_JOB_ID', '')} [/bold blue]!")
        console.print(f"[bold blue] TIMESTAMP {datetime.now(timezone.utc).strftime('%d/%m/%Y %H:%M:%S')} [/bold blue]!")

    @rank_zero_only
    def print_benchmark_profiler_report(
        self,
        speed_metrics_df: pd.DataFrame | None = None,
        time_metrics_df: pd.DataFrame | None = None,
        memory_metrics_df: pd.DataFrame | None = None,
        system_metrics_df: pd.DataFrame | None = None,
        model_summary: str | None = None,
    ) -> None:
        self.print_title()
        self.print_metadata()

        if time_metrics_df is not None:
            warnings.warn(
                "INFO: Time Report metrics represent single-node metrics (not multi-node aggregated metrics)",
            )
            warnings.warn("INFO: Metrics with a * symbol, represent the value after aggregating all steps")
            self.print_report("Time Profiling", time_metrics_df, color="green", emoji="alarm_clock")

        if speed_metrics_df is not None:
            warnings.warn(
                "INFO: Speed Report metrics are single-node metrics (not multi-node aggregated metrics)",
            )
            self.print_report("Speed Profiling", speed_metrics_df, color="yellow", emoji="racing_car")

        if memory_metrics_df is not None:
            warnings.warn("INFO: Memory Report metrics represent metrics aggregated across all nodes")
            self.print_report("Memory Profiling", memory_metrics_df, color="purple", emoji="floppy_disk")

        if system_metrics_df is not None:
            self.print_report("System Profiling", system_metrics_df, color="Red", emoji="desktop_computer")

        if model_summary is not None:
            self.print_report("Model Summary", model_summary, color="Orange", emoji="robot")
            num_gpus = self.config.hardware.num_gpus_per_node * self.config.hardware.num_nodes
            if num_gpus > 1:
                LOGGER.info("Model Summary displays the stats for a single GPU.")

    @staticmethod
    def write_benchmark_profiler_report() -> None:
        console.save_html("report.html")

    @staticmethod
    def to_df(sample_dict: dict[str, float], precision: str = ".5") -> pd.DataFrame:
        df = pd.DataFrame(sample_dict.items())
        df.columns = ["metric", "value"]
        df.value = df.value.apply(lambda x: f"%{precision}f" % x)
        return df

    @cached_property
    @rank_zero_only
    def speed_profile(self) -> None:
        """Speed profiler Report.

        Get speed metrics from Progress Bar for training and validation.
        """
        if self.config.diagnostics.benchmark_profiler.speed.enabled:
            # Find the first ProfilerProgressBar callback.
            for callback in self.callbacks:
                if isinstance(callback, ProfilerProgressBar):
                    return self.profiler.get_speed_profiler_df(callback)
            else:
                error_msg = "No ProfilerProgressBar callback found."
                raise ValueError(error_msg)
        else:
            return None

    def _get_logger(self) -> dict[str, Logger]:
        if (self.config.diagnostics.log.wandb.enabled) and (not self.config.diagnostics.log.wandb.offline):
            logger_info = {"logger_name": "wandb", "logger": self.wandb_logger}
        elif self.config.diagnostics.log.tensorboard.enabled:
            logger_info = {"logger_name": "tensorboard", "logger": self.tensorboard_logger}
        elif self.config.diagnostics.log.mlflow.enabled:
            logger_info = {"logger_name": "mlflow", "logger": self.mlflow_logger}
        else:
            LOGGER.warning("No logger enabled for system profiler")
            logger_info = None
        return logger_info

    @cached_property
    @rank_zero_only
    def system_profile(self) -> None:
        """System Profiler Report."""
        if self.config.diagnostics.benchmark_profiler.system.enabled:
            logger_info = self._get_logger()
            if logger_info:
                return self.profiler.get_system_profiler_df(
                    logger_name=logger_info["logger_name"],
                    logger=logger_info["logger"],
                )
            LOGGER.warning("System Profiler Report is not available")
            return None
        return None

    @cached_property
    @rank_zero_only
    def memory_profile(self) -> None:
        """Memory Profiler Report."""
        if self.config.diagnostics.benchmark_profiler.memory.enabled:
            return self.profiler.get_memory_profiler_df()
        return None

    @cached_property
    @rank_zero_only
    def time_profile(self) -> None:
        """Time Profiler Report."""
        if self.config.diagnostics.benchmark_profiler.time.enabled:
            return self.profiler.get_time_profiler_df()
        return None

    @cached_property
    def model_summary(self) -> str:
        if self.config.diagnostics.benchmark_profiler.model_summary.enabled:
            model = self.model
            return self.profiler.get_model_summary(model=model, example_input_array=self.example_input_array)
        return None

    @rank_zero_only
    def export_to_logger(self) -> None:
        if (self.config.diagnostics.log.wandb.enabled) and (not self.config.diagnostics.log.wandb.offline):
            self.to_wandb()

        elif self.config.diagnostics.log.mlflow.enabled:
            self.to_mlflow()

    def report(self) -> str:
        """Print report to console."""
        LOGGER.info("Generating Profiler reports")
        self.print_benchmark_profiler_report(
            memory_metrics_df=self.memory_profile,
            time_metrics_df=self.time_profile,
            speed_metrics_df=self.speed_profile,  # speed profile needs to be generated after time and memory reports
            system_metrics_df=self.system_profile,
            model_summary=self.model_summary,
        )

    def _get_extra_files(self) -> None:
        extra_files = []
        extra_files.extend(self.profiler.dirpath.glob("*.pickle"))
        # These trace files are too big to push to MLFlow so
        # we won't push them as artifacts extra_files.extend(self.profiler.dirpath.glob("*.json"))
        return extra_files

    def _log_reports_to_mlflow(self, run_id: str, data: pd.DataFrame, artifact_file: str, report_fname: str) -> None:
        self.mlflow_logger.experiment.log_table(
            run_id=run_id,
            data=data,
            artifact_file=artifact_file,
        )

        self.mlflow_logger.experiment.log_artifact(run_id, report_fname)

    @rank_zero_only
    def to_mlflow(self) -> None:
        """Log report into MLFlow."""
        LOGGER.info("logging to MLFlow Profiler report")
        self.write_benchmark_profiler_report()
        # check this https://stackoverflow.com/questions/71151054/how-to-log- d da-table-of-metrics-into-mlflow

        run_id = self.mlflow_logger.run_id
        if self.config.diagnostics.benchmark_profiler.system.enabled:
            self._log_reports_to_mlflow(
                run_id=run_id,
                data=self.system_profile,
                artifact_file="system_metrics_report.json",
                report_fname=self.profiler.system_report_fname,
            )

        if self.config.diagnostics.benchmark_profiler.time.enabled:
            self._log_reports_to_mlflow(
                run_id=run_id,
                data=self.time_profile,
                artifact_file="time_metrics_reports.json",
                report_fname=self.profiler.time_report_fname,
            )

        if self.config.diagnostics.benchmark_profiler.speed.enabled:
            self._log_reports_to_mlflow(
                run_id=run_id,
                data=self.speed_profile,
                artifact_file="speed_metrics_reports.json",
                report_fname=self.profiler.speed_report_fname,
            )

        if self.config.diagnostics.benchmark_profiler.memory.enabled:
            self._log_reports_to_mlflow(
                run_id=run_id,
                data=self.memory_profile,
                artifact_file="memory_metrics_reports.json",
                report_fname=self.profiler.memory_report_fname,
            )

            extra_files = self._get_extra_files()
            for file in extra_files:
                artifact_path = self.profiler.dirpath / file
                if artifact_path.is_file():
                    self.mlflow_logger.experiment.log_artifact(run_id, artifact_path)

        if self.config.diagnostics.benchmark_profiler.model_summary.enabled:
            self.mlflow_logger.experiment.log_artifact(run_id, self.profiler.model_summary_fname)

    @rank_zero_only
    def to_wandb(self) -> None:
        """Log report into  W&B."""
        LOGGER.info("logging to W&B Profiler report")
        self.write_benchmark_profiler_report()
        import wandb
        from pytorch_lightning.loggers.wandb import WandbLogger

        logger = WandbLogger(
            project=self.run_dict["project"],
            entity=self.run_dict["entity"],
            id=self.run_dict["id"],
            offline=self.config.diagnostics.log.wandb.offline,
            resume=self.run_dict["id"],
        )

        logger.experiment.log({"speed_metrics_report": wandb.Table(dataframe=self.speed_profile)})
        logger.experiment.log({"memory_metrics_report": wandb.Table(dataframe=self.system_profile)})
        logger.experiment.log({"time_metrics_report": wandb.Table(dataframe=self.time_profile)})
        logger.experiment.log({"memory_metrics_report": wandb.Table(dataframe=self.memory_profile)})
        logger.experiment.log({"model_summary_report": wandb.Table(dataframe=self.model_summary)})
        with Path("report.html").open("w") as f:
            logger.experiment.log({"reports_benchmark_profiler": wandb.Html(f)})
        logger.experiment.finish()

    @cached_property
    def callbacks(self) -> list[pl.callbacks.Callback]:
        callbacks = super().callbacks
        callbacks.append(ProfilerProgressBar())
        if self.config.diagnostics.benchmark_profiler.snapshot.enabled:
            from anemoi.training.diagnostics.callbacks.profiler import MemorySnapshotRecorder
            from anemoi.training.diagnostics.profilers import check_torch_version

            available = check_torch_version()

            if available:  # if torch is below 2.1.0, the callback will not be added
                callbacks.append(MemorySnapshotRecorder(self.config))
        return callbacks

    @cached_property
    def datamodule(self) -> AnemoiDatasetsDataModule:
        datamodule = super().datamodule
        # to generate a model summary with shapes we need a sample input array
        batch = next(iter(datamodule.train_dataloader()))
        if type(batch) in [list, tuple]:
            batch = batch[0]
        self.example_input_array = batch[
            :,
            0 : self.config.training.multistep_input,
            ...,
            self.data_indices.data.input.full,
        ]
        # If the input batch is sharded, replicate it to its full size
        if self.config.dataloader.read_group_size > 1:
            self.example_input_array = self.example_input_array.repeat(
                1,
                1,
                1,
                self.config.dataloader.read_group_size,
                1,
            )
        return datamodule

    @cached_property
    def profiler(self) -> BenchmarkProfiler:
        return BenchmarkProfiler(self.config)

    def _update_paths(self) -> None:
        """Update the paths in the configuration."""
        super()._update_paths()

        if self.run_id:  # when using mlflow only rank0 will have a run_id except when resuming runs
            # Multi-gpu new runs or forked runs - only rank 0
            # Multi-gpu resumed runs - all ranks
            self.config.hardware.paths.profiler = Path(self.config.hardware.paths.profiler, self.run_id)
        elif self.config.training.fork_run_id:
            parent_run = self.config.training.fork_run_id
            self.config.hardware.paths.profiler = Path(self.config.hardware.paths.profiler, parent_run)
        LOGGER.info("Profiler path: %s", self.config.hardware.paths.profiler)

    def _close_logger(self) -> None:
        if (self.config.diagnostics.log.wandb.enabled) and (not self.config.diagnostics.log.wandb.offline):
            # We need to close the W&B logger to be able to read the System Metrics
            self.wandb_logger.experiment.finish()

    def profile(self) -> None:
        """Profile the model."""
        self.train()
        self.report()
        self.export_to_logger()


@hydra.main(version_base=None, config_path="../config", config_name="config")
def main(config: DictConfig) -> None:
    AnemoiProfiler(config).profile()


if __name__ == "__main__":
    main()
